'''
Widgets for the meta data tab.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import re
import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, QPoint, pyqtSlot
from PyQt5.QtGui import QIcon, QFont, QFontMetrics
from PyQt5.QtWidgets import QDialogButtonBox
from .. import sqlitedb
from ..tools import getHLine, getXExpandYMinSizePolicy, parseAuthors,\
        getXExpandYExpandSizePolicy
from .. import _crossref

LOGGER=logging.getLogger(__name__)



class AdjustableTextEdit(QtWidgets.QTextEdit):

    edited_signal=pyqtSignal(str)  # field name
    def __init__(self,field,parent=None):
        '''
        Args:
            field (str): field name of this textedit, e.g. title, year ...
            parent (QWidget): parent widget.

        This modified QTextEdit doesn't show scroll bar, but adjusts height
        acorrding to contents and width.
        '''

        super(AdjustableTextEdit,self).__init__(parent)

        self.field=field # field name, e.g. title, year, tags_l ...
        self.fm=QFontMetrics(self.font())
        # pop up tooltip
        self.tooltip_label=QtWidgets.QLabel()
        self.tooltip_label.setWindowFlags(Qt.SplashScreen)
        self.tooltip_label.setMargin(3)
        self.tooltip_label.setStyleSheet('''
                background-color: rgb(235,225,120)
                ''')
        self.tooltip_text=''
        self.label_enabled=False

        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.textChanged.connect(self.resizeTextEdit)
        self.document().documentLayout().documentSizeChanged.connect(
                self.resizeTextEdit)
        self.setTabChangesFocus(True)


    def focusInEvent(self,event):

        #-----------------Pop up tool tip-----------------
        if self.label_enabled and self.tooltip_text:
            self.tooltip_label.move(self.mapToGlobal(
                QPoint(0, self.height()-120)))
            self.tooltip_label.setText(self.tooltip_text)
            self.tooltip_label.show()

        super(AdjustableTextEdit,self).focusInEvent(event)


    def focusOutEvent(self,event):

        #------------------Close tool tip------------------
        if self.document().isModified():
            self.edited_signal.emit(self.field)
        if self.label_enabled:
            self.tooltip_label.close()
        super(AdjustableTextEdit,self).focusOutEvent(event)


    def setText(self,text):

        super(AdjustableTextEdit,self).setText(text)
        self.document().setModified(False)


    def resizeTextEdit(self):
        '''Resize edit'''

        docheight=self.document().size().height()
        margin=self.document().documentMargin()
        self.setMinimumHeight(docheight+2*margin)
        self.setMaximumHeight(docheight+2*margin)

        return



class AdjustableTextEditWithFold(AdjustableTextEdit):

    fold_change_signal=pyqtSignal(str,bool)   # field, isfold
    def __init__(self,field,parent=None):
        '''
        Args:
            field (str): field name of this textedit, e.g. title, year ...
            parent (QWidget): parent widget.

        This modified QTextEdit add a fold button next to the edit, and fold/
        unfold long texts.
        '''

        super(AdjustableTextEditWithFold,self).__init__(parent)

        self.field=field  # why I can't remove this?
        self.is_fold=False
        self.fold_above_nl=3

        self.fold_button=QtWidgets.QToolButton()
        self.fold_button.setArrowType(Qt.DownArrow)
        font_height=self.fontMetrics().height()
        self.fold_button.setFixedWidth(int(font_height))
        self.fold_button.setFixedHeight(int(font_height))
        self.fold_button.clicked.connect(self.toggleFold)
        self.fold_button.setStyleSheet('''
        QToolButton {
            border: 1px solid rgb(190,190,190);
            background-color: rgb(190,190,190);
            border-radius: %dpx;
            font: bold %dpx;
            color: white;
            }

        QToolButton:pressed {
            border-style: inset;
            }
        ''' %(int(font_height/2), max(1,font_height-2))
        )


    def getNumberOfLines(self):
        '''Compute the number of lines currently showing'''

        fm=self.fontMetrics()
        doc=self.document()
        docheight=doc.size().height()
        margin=doc.documentMargin()
        nlines=(docheight-2*margin)/fm.height()

        return nlines


    def resizeTextEdit(self):
        '''Resize and show/hide fold button'''

        if self.getNumberOfLines()<self.fold_above_nl:
            self.fold_button.setVisible(False)
            #self.unfoldText()
        else:
            self.fold_button.setVisible(True)
        if self.is_fold:
            self.foldText()
        else:
            self.unfoldText()

        return


    def toggleFold(self):

        self.unfoldText() if self.is_fold else self.foldText()
        self.fold_change_signal.emit(self.field,self.is_fold)

        return


    def foldText(self):
        '''Fold text to show only the 1st line'''

        nlines=self.getNumberOfLines()
        if nlines>=self.fold_above_nl:
            fontheight=self.fontMetrics().height()
            margin=self.document().documentMargin()
            self.setMinimumHeight(fontheight+2*margin)
            self.setMaximumHeight(fontheight+2*margin)
            self.is_fold=True
            self.fold_button.setArrowType(Qt.LeftArrow)

        return


    def unfoldText(self):
        '''Expand text to show all lines'''

        docheight=self.document().size().height()
        margin=self.document().documentMargin()
        self.setMinimumHeight(docheight+2*margin)
        self.setMaximumHeight(docheight+2*margin)
        self.is_fold=False
        self.fold_button.setArrowType(Qt.DownArrow)

        return



class FileLineEdit(QtWidgets.QLineEdit):
    def __init__(self,parent=None):
        '''
        Args:
            parent (QWidget): parent widget.

        This modified QLineEdit accepts a file path as its text, and displays
        an elided version of the file name part of the path.
        '''

        super(FileLineEdit,self).__init__(parent)
        self.fm=QFontMetrics(self.font())

    def setText(self,text,elide=True):

        self.full_text=text
        self.short_text=os.path.split(self.full_text)[1]

        # It seems that it requires a 20 pixel extra space
        super(FileLineEdit,self).setText(
             self.fm.elidedText(self.short_text,Qt.ElideRight,self.width()-20))

        return


    def text(self):

        return self.fm.elidedText(self.short_text,Qt.ElideRight,self.width()-20)


    def resizeEvent(self,event):

        super(QtWidgets.QLineEdit, self).resizeEvent(event)
        if hasattr(self,'full_text'):
            self.setText(self.full_text,elide=True)



class MetaTabScroll(QtWidgets.QScrollArea):

    meta_edited=pyqtSignal(list) # send field names
    update_by_doi_signal=pyqtSignal(sqlitedb.DocMeta)

    def __init__(self,settings,parent=None):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''

        super(MetaTabScroll,self).__init__(parent)

        self.settings=settings
        self.parent=parent

        self.label_color='color: rgb(0,0,140); background-color: rgb(235,235,240)'
        self.label_font=QFont('Serif',12,QFont.Bold)

        frame=QtWidgets.QWidget()
        frame.setStyleSheet('background-color:white')
        self.setWidgetResizable(True)
        self.setWidget(frame)
        # key: field name, consistent with the keys in DocMeta
        # value: textedit or lineedit
        # therefore this dict is a translated version of DocMeta (subset)
        self.fields_dict={}
        self.fold_dict={} # key: field name, value: is textedit folded

        #-------------------Add widgets-------------------
        self.v_layout=QtWidgets.QVBoxLayout()

        #--------------------Add title--------------------
        title_te=AdjustableTextEdit('title')
        title_te.setFrameStyle(QtWidgets.QFrame.NoFrame)
        title_te.setFont(self.settings.value('display/fonts/meta_title', QFont))
        self.fields_dict['title']=title_te
        self.v_layout.addWidget(title_te)

        self.v_layout.addWidget(getHLine(self))

        #-------------------Add authors-------------------
        self.createMultiLineField('Authors','authors_l','meta_authors')

        #-----Add journal, year, volume, issue, pages-----
        grid_layout=QtWidgets.QGridLayout()

        for fii in ['publication','year','volume','issue','pages',
                'citationkey']:
            self.createOneLineField(fii,fii,'meta_keywords',grid_layout)

        self.v_layout.addLayout(grid_layout)

        #---------------------Add tags---------------------
        self.createMultiLineField('Tags','tags_l','meta_keywords')

        #-------------------Add abstract-------------------
        self.createMultiLineField('Abstract','abstract','meta_keywords')

        #-------------------Add keywords-------------------
        self.createMultiLineField('Keywords','keywords_l',
                'meta_keywords')

        #-----------------Add catalog ids-----------------
        self.v_layout.addWidget(self.createLabel('Catalog IDs'))

        grid_layout=QtWidgets.QGridLayout()

        for fii in ['arxivId','doi','issn','pmid']:
            self.createOneLineField(fii,fii,'meta_keywords',grid_layout)

        #--------------Add doi search button--------------
        self.doi_search_button=QtWidgets.QPushButton()
        self.doi_search_button.setFixedSize(30,30)
        self.doi_search_button.setIcon(QIcon.fromTheme('edit-find'))
        self.doi_search_button.setStyleSheet('''
        QPushButton {
            border: 1px solid rgb(190,190,190);
            border-radius: 15px;
            color: white;
            }
        QPushButton:pressed {
            border-style: inset;
            }
        ''')
        self.doi_search_button.clicked.connect(self.doiSearchButtonClicked)

        grid_layout.addWidget(self.doi_search_button,2,2)

        self.v_layout.addLayout(grid_layout)

        #--------------------Add files--------------------
        self.fields_dict['files_l']=[]
        self.v_layout.addWidget(self.createLabel('Files'))
        self.file_insert_idx=self.v_layout.count()

        LOGGER.debug('NO of widgets in v_layout=%d' %self.v_layout.count())

        #---------------Add add file button---------------
        add_file_button=QtWidgets.QPushButton()
        add_file_button.setText('Add File...')
        add_file_button.setSizePolicy(getXExpandYMinSizePolicy())
        add_file_button.setStyleSheet('''
        QPushButton {
            background-color: rgb(220,220,220);
            border-radius: 2px;
            }

        QPushButton:pressed {
            border-style: inset;
            }
        ''')
        add_file_button.clicked.connect(self.addFileButtonClicked)
        self.v_layout.addWidget(add_file_button)

        self.v_layout.addStretch()
        frame.setLayout(self.v_layout)

        #-------------Connect focus out events-------------
        for kk,vv in self.fields_dict.items():
            if isinstance(vv,(list,tuple)):
                for vii in vv:
                    vii.edited_signal.connect(self.fieldEdited)
            else:
                vv.edited_signal.connect(self.fieldEdited)

        #------------------Set tab orders------------------
        field_keys=['title', 'authors_l', 'publication', 'year', 'volume',
                'issue', 'pages', 'citationkey', 'tags_l', 'abstract',
                'keywords_l', 'arxivId', 'doi', 'issn', 'pmid']
        for ii,kk in enumerate(field_keys[:-1]):
            w1=self.fields_dict[kk]
            w2=self.fields_dict[field_keys[ii+1]]
            self.setTabOrder(w1,w2)


    @pyqtSlot(str)
    def fieldEdited(self,field):
        '''Collect edited fields and send meta_edited signal

        Args:
            field (str): field name

        Now it appears a bit redundant.
        '''

        LOGGER.debug('Changed field = %s' %field)
        LOGGER.debug('meta_dict = %s' %self._meta_dict)
        self.meta_edited.emit([field,])

        return


    def createLabel(self,label):

        qlabel=QtWidgets.QLabel(label)
        qlabel.setStyleSheet(self.label_color)
        qlabel.setFont(self.label_font)

        return qlabel


    def createOneLineField(self, label, key, font_name, grid_layout):
        '''Create a label and a textedit and put in a row

        Args:
            label (str): text to display in the label.
            key (str): field name for the textedit, this is also the key in
                       the DocMeta dict.
            font_name (str): string specifying the font used in the textedit.
            grid_layout (QGridLayout): grid layout to add the label and textedit.
        '''

        te=AdjustableTextEdit(key)
        qlabel=QtWidgets.QLabel(label)
        qlabel.setStyleSheet(self.label_color)

        te.setFont(self.settings.value('display/fonts/%s' %font_name, QFont))

        rnow=grid_layout.rowCount()
        grid_layout.addWidget(qlabel,rnow,0,1,1)
        if key=='doi':
            # leave room for button
            grid_layout.addWidget(te,rnow,1,1,1)
        else:
            grid_layout.addWidget(te,rnow,1,1,2)
        self.fields_dict[key]=te

        return


    def createMultiLineField(self, label, key, font_name):
        '''Create a label and a foldable textedit in 2 rows

        Args:
            label (str): text to display in the label.
            key (str): field name for the textedit, this is also the key in
                       the DocMeta dict.
            font_name (str): string specifying the font used in the textedit.
        '''

        te=AdjustableTextEditWithFold(key)
        te.setFrameStyle(QtWidgets.QFrame.NoFrame)
        self.fold_dict[key]=te.is_fold
        te.fold_change_signal.connect(self.foldChanged)

        te.setFont(self.settings.value('display/fonts/%s' %font_name, QFont))

        # set tooltip texts
        if key=='authors_l':
            te.label_enabled=True
            te.tooltip_text='lastname, firstname\nlastname, firstname\n...'
        elif key=='tags_l':
            te.label_enabled=True
            te.tooltip_text='tag1; tag2; tag3 ...'
        elif key=='keywords_l':
            te.label_enabled=True
            te.tooltip_text='keyword1; keyword2; keyword3 ...'
        self.fields_dict[key]=te

        h_layout=QtWidgets.QHBoxLayout()
        h_layout.addWidget(self.createLabel(label))
        h_layout.addWidget(te.fold_button)

        self.v_layout.addLayout(h_layout)
        self.v_layout.addWidget(te)

        return te


    def createFileField(self,text=None,font_name='meta_keywords'):
        '''Create lineedit for an attachment file and a associated del button

        Kwargs:
            text (str or None): file path.
            font_name (str): string specifying the font used in the textedit.
        '''

        h_layout=QtWidgets.QHBoxLayout()

        le=FileLineEdit()
        le.setReadOnly(True)
        le.setFont(self.settings.value('display/fonts/%s' %font_name, QFont))

        if text is not None:
            le.setText(text)

        if le not in self.fields_dict['files_l']:
            self.fields_dict['files_l'].append(le)

        # create a del file button
        button=QtWidgets.QPushButton()
        font_height=le.fm.height()
        button.setFixedWidth(int(font_height))
        button.setFixedHeight(int(font_height))
        button.setText('\u2715')
        button.setStyleSheet('''
        QPushButton {
            border: 1px solid rgb(190,190,190);
            background-color: rgb(190,190,190);
            border-radius: %dpx;
            font: bold %dpx;
            color: white;
            text-align: center;
            padding-bottom: 2px;
            }

        QPushButton:pressed {
            border-style: inset;
            }
        ''' %(int(font_height/2), max(1,font_height-2))
        )
        button.clicked.connect(lambda: self.delFileButtonClicked(
            self.fields_dict['files_l'].index(le)))

        le.del_button=button
        h_layout.addWidget(le)
        h_layout.addWidget(button)

        LOGGER.debug('Insert file %s entry at %s' %(text,self.file_insert_idx))

        # keep a record of the current idx in the vertical layout
        self.v_layout.insertLayout(self.file_insert_idx,h_layout)
        self.file_insert_idx+=1

        return


    def delFileButtonClicked(self, idx=None):
        '''Delete an attachment file in response to del button click'''

        self.delFileField(idx)
        self.fieldEdited('files_l')

        return


    def delFileField(self,idx=None):
        '''Delete an attachment file or all files

        Kwargs:
            idx (int or None): if int, the index of the file in
                fields_dict['files_l'] list to delete. If None, delete all files.
        '''

        def delFile(le):
            self.v_layout.removeWidget(le.del_button)
            self.v_layout.removeWidget(le)
            le.deleteLater()
            le.del_button.deleteLater()
            # NOTE: you can't del a element in list if it is iterating
            #self.fields_dict['files_l'].remove(le)
            # keep a record of the current idx in the vertical layout
            self.file_insert_idx-=1

        if idx is None:
            for leii in self.fields_dict['files_l']:
                delFile(leii)
            self.fields_dict['files_l']=[]

        else:
            if idx in range(len(self.fields_dict['files_l'])):
                leii=self.fields_dict['files_l'][idx]
                delFile(leii)
                self.fields_dict['files_l'].remove(leii)

        return


    def addFileButtonClicked(self):
        '''Add an attachment file, in response to add file button click
        '''

        if hasattr(self.parent, '_current_doc') and self.parent._current_doc is None:
            return

        fname = QtWidgets.QFileDialog.getOpenFileName(self, 'Choose a PDF file',
         '',"PDF files (*.pdf);; All files (*)")[0]

        if fname:

            LOGGER.info('Add new file = %s' %fname)
            self.createFileField(fname)
            self.fieldEdited('files_l')

        return


    @pyqtSlot(str, bool)
    def foldChanged(self, field, isfold):
        '''Store fold state of a field

        Args:
            field (str): name of field whose fold state is changed.
            isfold (bool): textedit is folded or not.
        '''

        self.fold_dict[field]=isfold
        LOGGER.debug('Field = %s. isfold = %s' %(field, isfold))

        return


    def doiSearchButtonClicked(self):
        '''Query meta data via doi in response to doi button click
        '''

        doi_pattern=re.compile(r'(?:doi:)?\s?(10.[1-9][0-9]{3}/.*$)',
                re.DOTALL|re.UNICODE)

        doi=self.fields_dict['doi'].toPlainText()
        LOGGER.info('doi = %s' %doi)

        if len(doi)>0:
            match=doi_pattern.match(doi)
            LOGGER.debug('match = %s' %match)

            if match:
                rec,doi_dict=_crossref.fetchMetaByDOI(doi)
                if rec==1:
                    LOGGER.warning('Failed to fetch from doi.')

                    msg=QtWidgets.QMessageBox()
                    msg.resize(500,500)
                    msg.setIcon(QtWidgets.QMessageBox.Information)
                    msg.setWindowTitle('Error')
                    msg.setText('Oopsie.')
                    msg.setInformativeText('Failed to retrieve metadata from doi')
                    msg.exec_()

                    return

                meta_dict=_crossref.crossRefToMetaDict(doi_dict)
                LOGGER.debug('Got meta_dict from doi:' %meta_dict)
                LOGGER.debug('citationkey = %s' %meta_dict['citationkey'])

                self.exchangeMetaDict(meta_dict)

        return


    def exchangeMetaDict(self, new_dict):
        '''Update the meta data dict of current doc with info from doi query

        Args:
            new_dict (DocMeta): meta data dict created in doi query.

        This will compare the meta data from doi query with existing ones,
        if any difference found, update.
        '''

        docid=self.parent._current_doc
        if docid is None:
            return

        old_dict=self.parent.meta_dict[docid]
        fields=[]

        for kk,vv in old_dict.items():
            # don't update these fields
            if kk in ['id', 'read', 'favourite', 'added', 'files_l',
                    'folders_l', 'tags_l', 'deletionPending', 'notes',
                    'abstract' ]:
                new_dict[kk]=vv
            else:
                if vv!=new_dict[kk]:
                    fields.append(kk)

        if 'firstNames_l' in fields or 'lastName_l' in fields:
            fields.append('authors_l') # updateTableData monitors this
            # What's point of this?

        LOGGER.debug('Changed fields = %s' %fields)

        if len(fields)==0:
            return

        self.update_by_doi_signal.emit(new_dict)

        return


    @property
    def _meta_dict(self):
        '''Create a meta dict from the info in the lineedit/textedit widgets
        '''

        def parseToList(text):
            result=[]
            textlist=text.replace('\n',';').strip(';').split(';')
            for tii in textlist:
                tii=tii.strip()
                if len(tii)>0:
                    result.append(tii)
            return result

        result_dict=sqlitedb.DocMeta()
        for kk,vv in self.fields_dict.items():
            # field should be a list if key ends with '_l'
            if kk.endswith('_l'):
                if isinstance(vv,(tuple,list)):
                    values=[]
                    for vii in vv:
                        if isinstance(vii,QtWidgets.QLineEdit):
                            if kk=='files_l':
                                textii=vii.full_text
                            else:
                                textii=vii.text().strip()
                        elif isinstance(vii,QtWidgets.QTextEdit):
                            textii=vii.toPlaintext().strip()
                        if textii:
                            values.append(textii)
                    result_dict[kk]=values
                elif isinstance(vv,QtWidgets.QTextEdit):
                    if kk=='authors_l':
                        names=parseToList(vv.toPlainText())
                        firsts,lasts,_=parseAuthors(names)
                        result_dict['firstNames_l']=firsts
                        result_dict['lastName_l']=lasts
                    else:
                        result_dict[kk]=parseToList(vv.toPlainText())
                elif isinstance(vv,QtWidgets.QLineEdit):
                    result_dict[kk]=parseToList(vv.toText())
            # field should be a str
            else:
                if isinstance(vv,QtWidgets.QLineEdit):
                    values=vv.toText().strip()
                    result_dict[kk]=values or None
                elif isinstance(vv,QtWidgets.QTextEdit):
                    values=vv.toPlainText().strip()
                    result_dict[kk]=values or None

        return result_dict


class MetaDataEntryDialog(QtWidgets.QDialog):
    '''Previous as a dialog for manually adding a doc.
    NOT IN USE currently. Can't quite recall, maybe it's the doi updating
    part that is rather tricky.
    '''

    def __init__(self,settings,parent=None):
        super(MetaDataEntryDialog,self).__init__(parent)

        self.resize(500,700)
        self.setWindowTitle('Add New Entry')
        self.setWindowModality(Qt.ApplicationModal)

        v_layout=QtWidgets.QVBoxLayout()

        self.scroll=MetaTabScroll(settings,self)

        self.buttons=QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
            Qt.Horizontal, self)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        self.scroll.meta_edited.connect(self.checkOkButton)

        v_layout.addWidget(self.scroll)
        v_layout.addWidget(self.buttons)

        self.setLayout(v_layout)

        self.initialized=False
        self.empty_dict=sqlitedb.DocMeta()
        self.empty_dict.pop('added')
        self.empty_dict.pop('files_l')

    def checkOkButton(self):
        def checkDictChanged(d1,d2):
            for kk in d1.keys():
                if d1[kk]!=d2[kk]:
                    return True
            return False

        if checkDictChanged(self.empty_dict,self.scroll._meta_dict):
            self.buttons.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.buttons.button(QDialogButtonBox.Ok).setEnabled(False)

    def showEvent(self,e):
        if not self.initialized:
            self.buttons.button(QDialogButtonBox.Ok).setEnabled(False)
            self.initialized=True

        self.scroll.fields_dict['title'].setFocus()
        super(MetaDataEntryDialog,self).showEvent(e)
        return

    def exec_(self):
        ret=super(MetaDataEntryDialog,self).exec_()
        return ret, self.scroll._meta_dict


class NoteTextEdit(QtWidgets.QTextEdit):

    note_edited_signal=pyqtSignal()

    def __init__(self,settings,parent=None):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py

        A slightly moded QTextEdit to displaying and editing note.
        Will have to add more formatting features later.
        '''

        self.settings=settings

        super(NoteTextEdit,self).__init__(parent=parent)

        self.setFont(self.settings.value('display/fonts/notes',QFont))
        self.setSizePolicy(getXExpandYExpandSizePolicy())


    def focusOutEvent(self,event):
        if self.document().isModified():
            self.note_edited_signal.emit()
        super(NoteTextEdit,self).focusOutEvent(event)




import setuptools

with open('README.md', 'r') as fin:
    long_description=fin.read()

setuptools.setup(
        name='MeiTingTrunk',
        version='0.1.a1',
        author='Guangzhi XU',
        author_email='xugzhi1987@gmail.com',
        description='An open source reference management tool developed in PyQt5 and Python3.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/Xunius/MeiTingTrunk',
        packages=setuptools.find_packages(),
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: X11 Applications :: Qt',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Programming Language :: Python :: 3',
            'Topic :: Education'
            ],
        install_requires=[
            'qt>=5',
            'pyqt>=5',
            'pdfminer',
            'pypdf2',
            'beautifulsoup4',
            'fuzzywuzzy',
            'bibtexparser',
            'crossrefapi',
            'RISparser'
            ],
        python_requires='>=3',
        #package_data={'sample': ['
        entry_points={
            'gui_scripts': ['main = MeiTingTrunk.main:main']
            }
        )



'''
DOI query.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import logging
from crossref.restful import Works, Etiquette
#from habanero import Crossref, cn
try:
    from . import sqlitedb
except:
    import sqlitedb
#from ..gui import __version__

__version__='v0.1alpha'


ETIQUETTE=Etiquette('MeiTing-Trunk', __version__, 'github',
        'xugzhi1987@gmail.com')

LOGGER=logging.getLogger(__name__)



def fetchMetaByDOI(doi):

    works=Works(etiquette=ETIQUETTE)
    try:
        data=works.doi(doi)
    except:
        rec=1

    if data is None:
        rec=1
    else:
        rec=0

    LOGGER.info('DOI = %s. Rec of doi query = %s' %(doi, rec))

    return rec,data




def crossRefToMetaDict(in_dict):

    result=sqlitedb.DocMeta()
    in_dict=dict([(kk.lower(),vv) for kk,vv in in_dict.items()])

    keys=['title', 'issue', 'pages', 'publication', 'volume', 'year', 'doi',
            'abstract', 'arxivId', 'chapter', 'city', 'country', 'edition',
            'institution', 'isbn', 'issn', 'month', 'day', 'publisher',
            'series', 'pmid' ]

    # get type
    _type=in_dict.get('type',None)
    if _type=='journal-article':
        result['type']='article'
    else:
        # not supported yet
        result['type']='article'

    # get authors
    authors=in_dict.get('author',None)
    if authors:
        firstnames=[aii.get('given','') for aii in authors]
        lastnames=[aii.get('family','') for aii in authors]
        result['firstNames_l']=firstnames
        result['lastName_l']=lastnames

    # get others
    for keyii in keys:
        if keyii in in_dict:
            vii=in_dict[keyii]
            if isinstance(vii, list):
                vii=''.join(vii) # title is a list
            result[keyii]=vii

    # sometimes this is journal name
    if 'journal-title' in in_dict:
        result['publication']=in_dict['journal-title']
    else:
        if 'container-title' in in_dict:
            vii=in_dict['container-title']
            if isinstance(vii, list):
                vii=''.join(vii) # title is a list
            result['publication']=vii

    if 'page' in in_dict:
        result['pages']=in_dict['page']

    if 'url' in in_dict:
        result['urls_l']=[in_dict['url'],]

    # get time
    if 'year' not in in_dict:
        if 'issued' in in_dict:
            issued=in_dict['issued']
        elif 'published-print' in in_dict:
            issued=in_dict['published-print']

        if 'date-parts' in issued:
            try:
                date=issued['date-parts'][0]
                if len(date)==2:
                    year,month=date
                    result['year']=str(year)
                    result['month']=str(month)
                elif len(date)==3:
                    year,month,day=date
                    result['year']=str(year)
                    result['month']=str(month)
                    result['day']=str(day)
            except:
                pass

    #result['confirmed']='true'

    return result



'''
Widget for the doc table: a table model and a headerview.

MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from datetime import datetime
import operator
import logging
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import QAbstractTableModel, Qt, QVariant, pyqtSignal,\
        pyqtSlot, QMimeData, QByteArray
from PyQt5.QtGui import QPixmap, QBrush, QColor, QIcon, QFont


LOGGER=logging.getLogger(__name__)



class TableModel(QAbstractTableModel):
    def __init__(self, parent, datain, headerdata, settings):
        '''
        Args:
            parent (QWidget): parent widget.
            datain (list): data for the table, a list of lists, each
                           element list for a row. Created by
                           _MainFrameLoadData.prepareDocs().
            headerdata (list): table column names.
            settings (QSettings): application settings. See _MainWindow.py
        '''

        QAbstractTableModel.__init__(self, parent)

        self.ncol=len(headerdata)
        if datain is None:
            self.arraydata=[None]*self.ncol
        else:
            self.arraydata=datain
        self.headerdata=headerdata
        self.settings=settings

        self.icon_section={
                'has_file': QIcon(':/file_icon.png')
                }
        self.check_section={
                'favourite': QPixmap(':/bf.png'),
                'read': QPixmap(':/clap.png')
                }
        self.icon_sec_indices=[self.headerdata.index(kk) for kk
                in self.icon_section.keys()]
        self.check_sec_indices=[self.headerdata.index(kk) for kk
                in self.check_section.keys()]


    def rowCount(self,p):
        return len(self.arraydata)


    def columnCount(self,p):
        return self.ncol


    def data(self, index, role):
        if not index.isValid():
            return QVariant()

        #if role == Qt.BackgroundRole:
            #if index.row()%2==0:
                #return QBrush(QColor(230,230,249))
                #pass

        if role == Qt.FontRole:
            font=self.settings.value('display/fonts/doc_table',QFont)
            if self.arraydata[index.row()][9] in [None, 'false']:
                font.setBold(True)
            else:
                font.setBold(False)
            return font

        if role==Qt.DisplayRole:
            if index.column() in self.icon_sec_indices:
                return
            elif index.column()==self.headerdata.index('added'):
                added=self.arraydata[index.row()][index.column()]
                if added:
                    # convert time to str
                    added=int(added[:10])
                    added=datetime.fromtimestamp(added)
                    if added.year==datetime.today().year:
                        added=added.strftime('%b-%d')
                    else:
                        added=added.strftime('%b-%d-%y')
                    return QVariant(added)
                else:
                    return
            else:
                return QVariant(self.arraydata[index.row()][index.column()])

        if role==Qt.EditRole:
            return QVariant(self.arraydata[index.row()][index.column()])

        #if role==Qt.TextAlignmentRole:
            #return Qt.AlignCenter
        if index.column() in self.check_sec_indices and role==Qt.CheckStateRole:
            if self.arraydata[index.row()][index.column()].isChecked():
                return Qt.Checked
            else:
                return Qt.Unchecked
        if index.column() in self.icon_sec_indices and role==Qt.DecorationRole:
            if self.arraydata[index.row()][index.column()]:
                return self.icon_section['has_file']
            else:
                return None

        if role != Qt.DisplayRole:
            return QVariant()


    def setData(self, index, value, role):
        if not index.isValid():
            return False
        if index.column() in self.check_sec_indices and role==Qt.CheckStateRole:
            if value == Qt.Checked:
                self.arraydata[index.row()][index.column()].setChecked(True)
            else:
                self.arraydata[index.row()][index.column()].setChecked(False)
        self.dataChanged.emit(index,index)

        return True


    def flags(self, index):
        if index.column() in self.check_sec_indices:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable\
                    | QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsDragEnabled
        else:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable |\
                    QtCore.Qt.ItemIsDragEnabled


    def headerData(self, col, orientation, role):

        if col in self.icon_sec_indices:
            label=self.headerdata[col]
            if orientation==Qt.Horizontal and role==Qt.DecorationRole:
                return self.icon_section[label]
        elif col in self.check_sec_indices:
            label=self.headerdata[col]
            if orientation==Qt.Horizontal and role==Qt.DecorationRole:
                return self.check_section[label]
        else:
            if orientation==Qt.Horizontal and role==Qt.DisplayRole:
                return self.headerdata[col]

        return None


    def sort(self,col,order):
        self.layoutAboutToBeChanged.emit()

        #NOTE that python3 doesn't support mixed type sorting (e.g. 1<None,
        # 'a' > 2. So convert everything to str.
        self.arraydata=sorted(self.arraydata,key=lambda x: \
                str(operator.itemgetter(col)(x)) or '')
        if order==Qt.DescendingOrder:
            self.arraydata.reverse()
        self.layoutChanged.emit()

        return


    def mimeTypes(self):
        '''For drag/drop docs into folders'''

        return ['doc_table_item',]


    def mimeData(self,indices):
        '''For drag/drop docs into folders'''

        LOGGER.debug('header data = %s' %self.headerdata)
        LOGGER.debug('indices = %s' %indices)

        for idii in indices:
            LOGGER.debug('idii.row() = %s, idii.column() = %s, idii.data() = %s'\
                    %(idii.row(), idii.column(), idii.data()))

        ids=[ii for ii in indices if ii.isValid()]
        LOGGER.debug('ids = %s' %ids)

        rowids=[ii.row() for ii in ids]
        rowids=list(set(rowids))
        LOGGER.debug('rowids = %s' %rowids)

        encode_data=[str(self.arraydata[ii][0]) for ii in rowids]
        encode_data=', '.join(encode_data)
        LOGGER.debug('encode_data = %s, type() = %s' %(encode_data,
            type(encode_data)))

        encode_data_array=QByteArray()
        encode_data_array.append(encode_data)

        mimedata=QMimeData()
        mimedata.setData('doc_table_item',encode_data_array)

        return mimedata


class MyHeaderView(QtWidgets.QHeaderView):
    def __init__(self,parent):
        super(MyHeaderView,self).__init__(Qt.Horizontal,parent)

        self.colSizes={'docid':0, 'favourite': 20, 'read': 20, 'has_file': 20,
            'author': 200, 'title': 500, 'journal':100,'year':50,'added':50,
            'confirmed':0}

        self.setSectionsClickable(True)
        self.setHighlightSections(True)
        self.sectionResized.connect(self.myresize)
        self.setStretchLastSection(False)
        self.setSectionsMovable(True)

    def initresizeSections(self):
        '''Initial resize columns

        The entire resizing logic is very confusing, I can't quite recall
        myself.

        The problem was: by default Qt doesn't let you resize each column's
        width, if you want the table widget to expand with its container.
        This is because there are only 4 resize modes for the headerview:
            * QHeaderView.Interactive
            * QHeaderView.Fixed
            * QHeaderView.Stretch
            * QHeaderView.ResizeToContents
        and they are mutually exclusive, i.e. you can't combine them like

            QHeaderView.Interactive | QHeaderView.Stretch

        Very annoying.
        The current solution is still not ideal.
        '''

        model=self.model()
        if model is None:
            return

        self.setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        headers=model.headerdata

        for ii in range(self.count()):
            lii=headers[ii]
            sii=self.colSizes[lii]
            if lii in self.colSizes:
                self.setSectionResizeMode(ii, QtWidgets.QHeaderView.Fixed)
                self.resizeSection(ii,sii)
            else:
                self.setSectionResizeMode(ii, QtWidgets.QHeaderView.Stretch)
                wnow=self.sectionSize(ii)
                self.resizeSection(ii,wnow)
                self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Interactive)

        return


    def myresize(self, *args):
        '''Resize columns
        '''

        model=self.model()
        if model is None:
            return
        ws=[]
        for c in range(self.count()):
            wii=self.sectionSize(c)
            ws.append(wii)

        if args[0]>0 or args[0]<self.count():
            for ii in range(args[0],self.count()):
                lii=model.headerdata[ii]
                if lii in ['favourite','read','has_file']:
                    continue
                if ii==args[0]:
                    continue
                if ii==self.count()-1:
                    self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Stretch)
                else:
                    pass

        return


    def resizeEvent(self, event):

        super(QtWidgets.QHeaderView, self).resizeEvent(event)

        model=self.model()
        if model is None:
            return

        ws=[]
        perc=[]
        total_w=self.length() # width of the table
        total_w2=self.size().width()   # new available space after resizing
        for c in range(self.count()):
            wii=self.sectionSize(c)
            ws.append(wii)
            perc.append(float(wii)/total_w)

        headers=model.headerdata

        for ii in range(self.count()):
            lii=headers[ii]
            if lii in ['favourite','read','has_file']:
                self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Fixed)
                continue
            elif lii=='year':
                self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Stretch)
                self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Interactive)
            else:
                wnow=int(perc[ii]*total_w2)
                self.resizeSection(ii,wnow)
                self.setSectionResizeMode(ii,QtWidgets.QHeaderView.Interactive)

        return


    def columnFromLabel(self, label):
        '''NOT IN USE'''
        headers=self.model().headerdata
        if label in headers:
            return headers.index(label)

        return -1




'''
Parse bibtex files into dict format, and export DocMeta dict to bibtex file.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import re
import logging
from pprint import pprint
import bibtexparser
from bibtexparser.bparser import BibTexParser
from bibtexparser.bwriter import BibTexWriter
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser import customization as bibcus
try:
    from . import sqlitedb
except:
    import sqlitedb
try:
    from .tools import parseAuthors
except:
    from tools import parseAuthors

# replace dict keys
ALT_KEYS={
        'keyword': 'keywords_l',
        'tag': 'tags_l',
        'url': 'urls_l',
        'file': 'files_l',
        'folder': 'folders_l'
        }

INV_ALT_KEYS=dict([(vv,kk) for kk,vv in ALT_KEYS.items()])

# for testing only
OMIT_KEYS=[
        'read', 'favourite', 'added', 'confirmed', 'firstNames_l',
        'lastName_l', 'deletionPending', 'folders_l', 'type', 'id'
        ]

LOGGER=logging.getLogger(__name__)



def splitFields(record, key, sep=',|;'):
    """Split keyword field into a list

    Args:
        record (dict): record dict.
        key (str): key in record dict to split its value.

    Kwargs:
        sep (str): pattern used for the splitting regexp.

    Returns: record (dict): the modified record

    """
    if key in record:
        record[key] = [i.strip() for i in re.split(sep, record[key].replace('\n', ''))]

    return record


def getPublication(record):
    '''Use journal field and publication

    maybe just for ARTICLE type??
    '''
    if 'journal' in record and 'publication' not in record:
        record['publication']=record['journal']

    return record


def customizations(record):
    """Use some functions delivered by the library

    Args:
        record (dict): record dict.

    Returns: record (dict): the modified record
    """
    record = bibcus.type(record)
    record = bibcus.author(record)
    #record = bibcus.editor(record)
    #record = bibcus.journal(record)
    record = bibcus.keyword(record)
    #record = bibcus.link(record)
    record = bibcus.page_double_hyphen(record)
    #record = bibcus.doi(record)
    record = splitFields(record, 'folder')
    record = splitFields(record, 'url', '\n')
    record = splitFields(record, 'file', ',|;|\n')
    record = getPublication(record)
    return record


def splitNames(entry):
    """Split authors list into a list of firstnames and a list of lastnames

    Args:
        entry (dict): meta data dict.

    Returns: entry (dict): meta data dict with a new 'firstNames_l' and
                           'lastName_l' key
    """

    firstnames=[]
    lastnames=[]
    for nii in entry['author']:
        lii,fii=nii.split(',',1)
        firstnames.append(fii.strip())
        lastnames.append(lii.strip())

    entry['firstNames_l']=firstnames
    entry['lastName_l']=lastnames

    return entry


def altKeys(entry_dict, alt_dict):

    for kk in alt_dict.keys():
        if kk in entry_dict:
            entry_dict[alt_dict[kk]]=entry_dict[kk]
            del entry_dict[kk]

    return entry_dict


def readBibFile(bibfile):
    """Read and parse bibtex file.

    Args:
        bibfile (str): abspath to input bibtex file.

    Returns: results (list): DocMeta dicts, each for an parsed entry in the
                             bibtex file.
    """

    bibfile=os.path.abspath(bibfile)
    if not os.path.exists(bibfile):
        return None

    with open(bibfile,'r') as fin:
        parser=BibTexParser()
        parser.homogenize_fields=True
        parser.customization=customizations
        bib=bibtexparser.load(fin,parser=parser)

        LOGGER.info('Read in bib file: %s' %bibfile)

    results=[]

    for eii in bib.entries:
        eii=splitNames(eii)
        eii=altKeys(eii,ALT_KEYS)
        eii['citationkey']=eii['ID']

        if 'folders_l' in eii:
            del eii['folders_l']

        docii=sqlitedb.DocMeta()
        docii.update(eii)
        results.append(docii)

    return results


def toOrdinaryDict(metadict, alt_dict, omit_keys, path_prefix):
    """Convert a DocMeta dict to an ordinary dict for bibtex export

    Args:
        metadict (DocMeta): meta dict of a doc.
        alt_dict (dict): dict for key changes.
        omit_keys (list): keys to omit in the converted dict.
        path_prefix (str): folder path to prepend to attachment file paths.

    Returns: result (dict): standard python dict containing meta data copied
                            from input <metadict> with a format suitable
                            for bibtex export.
    """

    result={}

    for kk,vv in metadict.items():
        if kk in omit_keys:
            continue

        if vv is None:
            continue

        if kk in alt_dict:
            if isinstance(vv,(tuple,list)):
                if len(vv)==0:
                    continue
                if kk in ['folders_l',]:
                    vv=[vii[1] for vii in vv]
                if kk == 'files_l':
                    vv=[os.path.join(path_prefix,vii) for vii in vv]
                vv='; '.join(vv)
            result[alt_dict[kk]]=vv
        else:
            result[kk]=str(vv)

        #LOGGER.debug('key = %s, value = %s' %(str(kk), str(vv)))

    authors=parseAuthors(metadict['authors_l'])[2]
    authors=' and '.join(authors)
    result['author']=authors

    doctype=metadict['type']
    if doctype is None or doctype.lower()=='journalarticle':
        doctype='article'
    result['ENTRYTYPE']=doctype

    ID=metadict['citationkey'].replace('(','').replace(')','')
    result['ID']=ID

    #LOGGER.debug('authors = %s' %authors)
    #LOGGER.debug('ENTRYTYPE = %s' %doctype)
    #LOGGER.debug('ID = %s' %ID)

    return result


def metaDictToBib(jobid, metadict, omit_keys, path_prefix):
    """Export meta data to bibtex format

    Args:
        jobid (int): id of job.
        metadict (DocMeta): meta dict of a doc.
        alt_dict (dict): dict for key changes.
        omit_keys (list): keys to omit in the converted dict.
        path_prefix (str): folder path to prepend to attachment file paths.

    Returns:
        rec (int): 0 if successful, 1 otherwise.
        jobid (int): the input jobid as it is.
        dbtext (str): formated bibtex entry, '' if <rec>==1.
        docid (int): id of the processed document.
    """

    try:
        alt_dict=INV_ALT_KEYS
        ord_dict=toOrdinaryDict(metadict,alt_dict,omit_keys,path_prefix)

        db=BibDatabase()
        db.entries=[ord_dict,]
        writer=BibTexWriter()
        writer.indent='    '
        writer.comma_first=False
        dbtext=writer.write(db)

        return 0,jobid,dbtext,metadict['id']

    except Exception:
        LOGGER.exception('Failed to write to bibtex')
        return 1,jobid,'',metadict['id']




if __name__=='__main__':
    aa=readBibFile('test.bib')
    pprint(aa[-1])


    # test export

    entries=[]
    for eii in aa:
        dii=toOrdinaryDict(eii,INV_ALT_KEYS,OMIT_KEYS,'/home/')
        entries.append(dii)

    db=BibDatabase()
    db.entries=entries
    writer=BibTexWriter()
    writer.indent='    '
    writer.comma_first=False

    with open('testexport.bib','w') as fout:
        dbtext=writer.write(db)
        print('dbtext',dbtext)
        fout.write(dbtext)

